--[[
Interface: 1.7.1.0 b10490

Copyright (C) GtX (Andy), 2020

Author: GtX | Andy
Date: 24.08.2020

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

Important:
Not to be added to any mods / maps or modified from its current release form.
No changes are to be made to this script without permission from GtX | Andy

Darf nicht zu Mods / Maps hinzugefügt oder von der aktuellen Release-Form geändert werden.
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
]]


local loadError
local numInserted = 0

local versionString = "0.0.0.0"
local buildId = 3

local modName = g_currentModName
local modDirectory = g_currentModDirectory


local function loadMapFinished(mission)
    if numInserted > 0 and mission.hud ~= nil then
        local speedMeter = mission.hud.speedMeter

        speedMeter.updateSpeedGauge = Utils.overwrittenFunction(speedMeter.updateSpeedGauge, function(speedMeterDisplay, superFunc, dt)
            superFunc(speedMeterDisplay, dt)

            if speedMeterDisplay.vehicle.getSyncedSpeedActive ~= nil and speedMeterDisplay.vehicle:getSyncedSpeedActive() then
                local r, g, b, a = 0.0953, 1, 0.0685, 0.75

                if speedMeterDisplay.originalGaugeColour == nil then
                    speedMeterDisplay.originalGaugeColour = {speedMeterDisplay.gaugeBackgroundElement:getColor()}
                end

                -- If waiting for synced vehicle to start moving then pulse the gauge.
                if speedMeterDisplay.vehicle:getSyncedSpeedWaiting() and MathUtil.clamp(math.cos(8.0 * (g_currentMission.time / 1000)) + 0.2, 0, 1) == 0 then
                    r, g, b, a = 1, 1, 1, 1
                end

                speedMeterDisplay.gaugeBackgroundElement:setColor(r, g, b, a)
            else
                if speedMeterDisplay.originalGaugeColour ~= nil then
                    speedMeterDisplay.gaugeBackgroundElement:setColor(unpack(speedMeterDisplay.originalGaugeColour))
                    speedMeterDisplay.originalGaugeColour = nil
                end
            end
        end)

        -- Guidance Steering Fix
        if g_guidanceSteering ~= nil and g_guidanceSteering.modName ~= nil then
            local enviroment = _G[g_guidanceSteering.modName]

            if enviroment ~= nil and enviroment.DriveUtil ~= nil then
                enviroment.DriveUtil.accelerateInDirection = Utils.overwrittenFunction(enviroment.DriveUtil.accelerateInDirection, function(vehicle, superFunc, axisForward, dt, forceBrake)
                    local spec = vehicle.spec_vehicleSpeedSync

                    if spec ~= nil and spec.speedSyncActive then
                        if axisForward == 0 then
                            if vehicle:getVehicleIsInRange(spec.vehicleInRange, spec.setRange) then
                                local motor = vehicle:getMotor()

                                local speedLimit = motor:getSpeedLimit(true)
                                local lastSpeed = spec.vehicleInRange:getLastSpeed()

                                if lastSpeed > 0.5 and lastSpeed < speedLimit + 5 then
                                    axisForward = MathUtil.sign(spec.vehicleInRange.movingDirection or 1)
                                    motor:setSpeedLimit(math.min(lastSpeed, speedLimit))
                                    spec.isWaiting = false
                                elseif not spec.isWaiting then
                                    vehicle:setSyncedSpeedState(false, nil, true)
                                end
                            else
                                vehicle:setSyncedSpeedState(false, nil, true)
                            end
                        else
                            vehicle:setSyncedSpeedState(false)
                        end
                    end

                    return superFunc(vehicle, axisForward, dt, forceBrake)
                end)
            end
        end
    end
end

local function finalizeVehicleTypes(vehicleTypeManager)
    local specializationName = string.format("%s.vehicleSpeedSync", modName)
    local specializationObject = g_specializationManager:getSpecializationObjectByName(specializationName)

    if specializationObject ~= nil then
        for typeName, typeEntry in pairs(vehicleTypeManager.vehicleTypes) do
            if specializationObject.prerequisitesPresent(typeEntry.specializations) then
                g_vehicleTypeManager:addSpecialization(typeName, specializationName)
                numInserted = numInserted + 1
            end
        end
    end
end

local function loadSettingsFromXML(filepath)
    if fileExists(filepath) then
        local xmlFile = loadXMLFile("VehicleSpeedSyncXML", filepath)

        if xmlFile ~= nil and xmlFile ~= 0 then
            VehicleSpeedSync.SPEED_SYNC_VISIBLE = Utils.getNoNil(getXMLBool(xmlFile, "vehicleSpeedSync.activeSpeedSync#showInputBinding"), true)
            VehicleSpeedSync.CRUISE_SYNC_VISIBLE = Utils.getNoNil(getXMLBool(xmlFile, "vehicleSpeedSync.cruiseControlSync#showInputBinding"), true)

            VehicleSpeedSync.SPEED_SYNC_ACTIVE = Utils.getNoNil(getXMLBool(xmlFile, "vehicleSpeedSync.activeSpeedSync#isActive"), true)
            VehicleSpeedSync.CRUISE_SYNC_ACTIVE = Utils.getNoNil(getXMLBool(xmlFile, "vehicleSpeedSync.cruiseControlSync#isActive"), true)

            delete(xmlFile)
        end
    else
        local xmlFile = io.open (filepath, "w")

        if xmlFile ~= nil and xmlFile ~= 0 then
            xmlFile:write(string.format([[
<?xml version="1.0" encoding="utf-8" standalone="no" ?>

<vehicleSpeedSync build="%d" version="%s">
    <!-- Changes made here will affect all single and multiplayer games. -->

    <!-- If 'showInputBinding' is 'false' then input binding information will not be displayed in the F1 menu. -->
    <!-- If 'isActive' is 'false' the given feature will be disabled completely. -->

    <activeSpeedSync showInputBinding="true" isActive="true"/>
    <cruiseControlSync showInputBinding="true" isActive="true"/>
</vehicleSpeedSync>]], buildId, versionString))

            xmlFile:close()
        end
    end
end

local function init()
    local mod = g_modManager:getModByName(modName)
    versionString = tostring(mod.version) or versionString

    if mod.modName == "FS19_VehicleSpeedSync" or mod.modName == "FS19_VehicleSpeedSync_update" and mod.author:len() == 3 then
        source(modDirectory .. "scripts/specializations/events/VehicleSpeedSyncEvent.lua")
        g_specializationManager:addSpecialization("vehicleSpeedSync", "VehicleSpeedSync", modDirectory .. "scripts/specializations/VehicleSpeedSync.lua")

        FSBaseMission.loadMapFinished = Utils.prependedFunction(FSBaseMission.loadMapFinished, loadMapFinished)
        VehicleTypeManager.finalizeVehicleTypes = Utils.prependedFunction(VehicleTypeManager.finalizeVehicleTypes, finalizeVehicleTypes)

        loadSettingsFromXML(getUserProfileAppPath() .. "vehicleSpeedSync.xml")
    else
        loadError = {
            author = mod.author,
            startUpdateTime = 2400,

            update = function(self, dt)
                self.startUpdateTime = self.startUpdateTime - dt

                if self.startUpdateTime < 0 then
                    local title = string.format("Vehicle Speed Sync - Version %s", versionString)
                    local text = string.format(g_i18n:getText("VSS_loadError", modName), modName, self.author)

                    if g_dedicatedServerInfo == nil then
                        if not g_gui:getIsGuiVisible() then
                            g_gui:showYesNoDialog({
                                title = title,
                                text = text,
                                dialogType = DialogElement.TYPE_LOADING,
                                callback = self.openModHubLink,
                                target = nil,
                                yesText = g_i18n:getText("button_ok"),
                                noText = g_i18n:getText("button_modHubDownload")
                            })
                        end
                    else
                        print("\n" .. text .. "\n    - https://farming-simulator.com/mods.php?lang=en&country=be&title=fs2019&filter=org&org_id=129652&page=0" .. "\n")
                        self.openModHubLink(true)
                    end
                end
            end,

            openModHubLink = function(ignore)
                if ignore == false then
                    local language = g_languageShort
                    local link = "mods.php?lang=en&country=be&title=fs2019&filter=org&org_id=129652&page=0"
                    if language == "de" or language == "fr" then
                        link = "mods.php?lang=" .. language .. "&country=be&title=fs2019&filter=org&org_id=129652&page=0"
                    end

                    openWebFile(link, "")
                end

                removeModEventListener(loadError)
                loadError = nil
            end
        }

        addModEventListener(loadError)
    end
end

init()
