--[[
Interface: 1.7.1.0 b10490

Copyright (C) GtX (Andy), 2020

Author: GtX | Andy
Date: 24.08.2020

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

Important:
Not to be added to any mods / maps or modified from its current release form.
No changes are to be made to this script without permission from GtX | Andy

Darf nicht zu Mods / Maps hinzugefügt oder von der aktuellen Release-Form geändert werden.
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
]]


VehicleSpeedSync = {}

VehicleSpeedSync.DEFAULT_RANGE = 25

VehicleSpeedSync.SPEED_SYNC_VISIBLE = true
VehicleSpeedSync.SPEED_SYNC_ACTIVE = true

VehicleSpeedSync.CRUISE_SYNC_VISIBLE = true
VehicleSpeedSync.CRUISE_SYNC_ACTIVE = true

VehicleSpeedSync.MOD_NAME = g_currentModName
VehicleSpeedSync.MOD_DIR = g_currentModDirectory

function VehicleSpeedSync.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Drivable, specializations) and not SpecializationUtil.hasSpecialization(Locomotive, specializations)
end

function VehicleSpeedSync.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "findVehicleInRange",    VehicleSpeedSync.findVehicleInRange)
    SpecializationUtil.registerFunction(vehicleType, "getVehicleIsInRange",   VehicleSpeedSync.getVehicleIsInRange)
    SpecializationUtil.registerFunction(vehicleType, "getCanSyncSpeed",       VehicleSpeedSync.getCanSyncSpeed)
    SpecializationUtil.registerFunction(vehicleType, "getSyncedVehicleSpeed", VehicleSpeedSync.getSyncedVehicleSpeed)
    SpecializationUtil.registerFunction(vehicleType, "setSyncedSpeedState",   VehicleSpeedSync.setSyncedSpeedState)
    SpecializationUtil.registerFunction(vehicleType, "getSyncedSpeedActive",  VehicleSpeedSync.getSyncedSpeedActive)
    SpecializationUtil.registerFunction(vehicleType, "getSyncedSpeedWaiting", VehicleSpeedSync.getSyncedSpeedWaiting)
end

function VehicleSpeedSync.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "updateVehiclePhysics",        VehicleSpeedSync.updateVehiclePhysics)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getCanStartAIVehicle",        VehicleSpeedSync.getCanStartAIVehicle)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "setCruiseControlState",       VehicleSpeedSync.setCruiseControlState)
    -- SpecializationUtil.registerOverwrittenFunction(vehicleType, "setCruiseControlMaxSpeed", VehicleSpeedSync.setCruiseControlMaxSpeed)
end

function VehicleSpeedSync.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad",                 VehicleSpeedSync)
    SpecializationUtil.registerEventListener(vehicleType, "onDelete",               VehicleSpeedSync)
    SpecializationUtil.registerEventListener(vehicleType, "onDraw",                 VehicleSpeedSync)
    SpecializationUtil.registerEventListener(vehicleType, "onLeaveVehicle",         VehicleSpeedSync)
    SpecializationUtil.registerEventListener(vehicleType, "onEnterVehicle",         VehicleSpeedSync)
    SpecializationUtil.registerEventListener(vehicleType, "onRegisterActionEvents", VehicleSpeedSync)
end

function VehicleSpeedSync:onLoad(savegame)
    self.spec_vehicleSpeedSync = self[string.format("spec_%s.vehicleSpeedSync", VehicleSpeedSync.MOD_NAME)]

    local spec = self.spec_vehicleSpeedSync

    spec.setRange = VehicleSpeedSync.DEFAULT_RANGE

    spec.vehicleInRange = nil
    spec.speedSyncActive = false
    spec.isWaiting = false

    if self.isClient then
        spec.vehicleName = self:getName()

        local xmlFile = loadXMLFile("SpeedSyncSounds", Utils.getFilename("sounds/speedSyncSounds.xml", VehicleSpeedSync.MOD_DIR))

        if xmlFile ~= nil then
            spec.samples = {
                activate = g_soundManager:loadSampleFromXML(xmlFile, "speedSyncSounds", "activate", VehicleSpeedSync.MOD_DIR, self.components, 1, AudioGroup.VEHICLE, nil, self),
                deactivate = g_soundManager:loadSampleFromXML(xmlFile, "speedSyncSounds", "deactivate", VehicleSpeedSync.MOD_DIR, self.components, 1, AudioGroup.VEHICLE, nil, self),
                connectionLost = g_soundManager:loadSampleFromXML(xmlFile, "speedSyncSounds", "connectionLost", VehicleSpeedSync.MOD_DIR, self.components, 1, AudioGroup.VEHICLE, nil, self)
            }

            delete(xmlFile)
        end
    end
end

function VehicleSpeedSync:onDelete()
    local spec = self.spec_vehicleSpeedSync

    self:setSyncedSpeedState(false, nil, false, true)

    if self.isClient ~= nil and spec ~= nil then
        for _, sample in pairs(spec.samples) do
            g_soundManager:deleteSample(sample)
        end
    end
end

function VehicleSpeedSync:onDraw(isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    if self.isClient and self.getIsEntered ~= nil and self:getIsEntered() then
        local spec = self.spec_vehicleSpeedSync

        if self:getSyncedSpeedActive() then
            local enterableSpec = spec.vehicleInRange.spec_enterable

            if enterableSpec ~= nil then
                local vehicleName, controllerName = nil, "AI"

                if not spec.vehicleInRange:getIsAIActive() then
                    if enterableSpec.playerStyle ~= nil and enterableSpec.playerStyle.playerName ~= nil then
                        controllerName = enterableSpec.playerStyle.playerName
                    end
                end

                if spec.vehicleInRange.spec_vehicleSpeedSync ~= nil then
                    vehicleName = spec.vehicleInRange.spec_vehicleSpeedSync.vehicleName
                end

                if vehicleName == nil then
                    vehicleName = spec.vehicleInRange:getName()
                end

                g_currentMission:addExtraPrintText(string.format("Speed Sync:  %s  (%s)", vehicleName, controllerName))
            end

            -- Stop the HUD flashing if vehicle starts to move
            if spec.isWaiting and spec.vehicleInRange ~= nil then
                if spec.vehicleInRange:getLastSpeed() > 1 then
                    spec.isWaiting = false
                end
            end
        end

        if VehicleDebug.state == VehicleDebug.DEBUG then
            local vehicle = spec.vehicleInRange

            if vehicle == nil or g_currentMission.interactiveVehicles[vehicle] == nil then
                vehicle = self:findVehicleInRange()
            end

            if vehicle ~= nil then
                local speedText = "Vehicle Speed: %.1f"
                local distance = calcDistanceFrom(self.rootNode, vehicle.rootNode) or 0
                local speed = math.max(0, vehicle:getLastSpeed() * vehicle.spec_motorized.speedDisplayScale)

                if speed < 0.5 then
                    speed = 0
                end

                if spec.isWaiting then
                    speedText = "Vehicle Speed: %.1f ( Waiting )"
                end

                g_currentMission:addExtraPrintText(string.format(speedText, speed))
                g_currentMission:addExtraPrintText(string.format("Vehicle Distance: %.2f", distance))
            end
        end
    end
end

function VehicleSpeedSync:onRegisterActionEvents(isActiveForInput, isActiveForInputIgnoreSelection)
    if self.isClient then
        local spec = self.spec_vehicleSpeedSync

        self:clearActionEventsTable(spec.actionEvents)
        spec.toggleSpeedSyncEventId = nil

        if self:getIsActiveForInput(true, false) and self.getIsEntered ~= nil and self:getIsEntered() then
            if Utils.getNoNil(VehicleSpeedSync.CRUISE_SYNC_ACTIVE, true) then
                local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.SYNC_CRUISE_CONTROL, self, VehicleSpeedSync.actionEventSyncCruiseControl, false, true, false, true)
                g_inputBinding:setActionEventText(actionEventId, g_i18n:getText("action_syncCruiseControl"), VehicleSpeedSync.MOD_NAME)
                g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_LOW)
                g_inputBinding:setActionEventTextVisibility(actionEventId, VehicleSpeedSync.CRUISE_SYNC_VISIBLE)
            end

            if Utils.getNoNil(VehicleSpeedSync.SPEED_SYNC_ACTIVE, true) then
                local _, actionEventId = self:addActionEvent(spec.actionEvents, InputAction.TOGGLE_SPEED_SYNC, self, VehicleSpeedSync.actionEventSpeedSyncState, false, true, false, true)
                g_inputBinding:setActionEventText(actionEventId, g_i18n:getText("action_activateSpeedSync"), VehicleSpeedSync.MOD_NAME)
                g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_LOW)
                g_inputBinding:setActionEventTextVisibility(actionEventId, VehicleSpeedSync.SPEED_SYNC_VISIBLE)

                spec.toggleSpeedSyncEventId = actionEventId
            end
        end
    end
end

function VehicleSpeedSync:onLeaveVehicle()
    self:setSyncedSpeedState(false, nil, false, true)
end

function VehicleSpeedSync:onEnterVehicle(isControlling)
    self:setSyncedSpeedState(false, nil, false, true)
end

function VehicleSpeedSync:findVehicleInRange(range)
    local vehicleInRange

    if self.isClient and self:getCanSyncSpeed() then
        local lastDistance = math.huge

        for _, vehicle in pairs (g_currentMission.interactiveVehicles) do
            if vehicle ~= self and vehicle.spec_drivable ~= nil and vehicle:getIsMotorStarted() then
                local distance = calcDistanceFrom(self.rootNode, vehicle.rootNode)

                if distance <= (range or VehicleSpeedSync.DEFAULT_RANGE) and distance < lastDistance then
                    lastDistance = distance
                    vehicleInRange = vehicle
                end
            end
        end
    end

    return vehicleInRange
end

function VehicleSpeedSync:getVehicleIsInRange(vehicle, range)
    if vehicle ~= nil and g_currentMission.interactiveVehicles[vehicle] ~= nil then
        return calcDistanceFrom(self.rootNode, vehicle.rootNode) <= (range or VehicleSpeedSync.DEFAULT_RANGE)
    end

    return false
end

function VehicleSpeedSync:getCanSyncSpeed()
    if self.getIsEntered ~= nil and self:getIsEntered() then
        return self:getIsActiveForInput(true, false) and not self:getIsAIActive()
    end

    return false
end

function VehicleSpeedSync:setSyncedSpeedState(active, vehicleInRange, connectionLost, noEventSend)
    local spec = self.spec_vehicleSpeedSync

    if spec ~= nil then
        active = Utils.getNoNil(active and vehicleInRange ~= nil, false)

        if spec.speedSyncActive ~= active then
            if active and self:getVehicleIsInRange(vehicleInRange, spec.setRange) then
                spec.vehicleInRange = vehicleInRange
                spec.speedSyncActive = true
                spec.isWaiting = vehicleInRange:getLastSpeed() <= 0.5
            else
                spec.speedSyncActive = false
                spec.vehicleInRange = nil
                spec.isWaiting = false
            end

            if noEventSend == nil or not noEventSend then
                if not self.isServer then
                    g_client:getServerConnection():sendEvent(SetVehicleSpeedSyncEvent:new(self, spec.speedSyncActive, spec.vehicleInRange, connectionLost))
                else
                    local owner = self:getOwner()

                    if owner ~= nil then
                        owner:sendEvent(SetVehicleSpeedSyncEvent:new(self, spec.speedSyncActive, spec.vehicleInRange, connectionLost))
                    end
                end
            end

            if self.isClient and self.getIsEntered ~= nil and self:getIsEntered() then
                if spec.speedSyncActive then
                    g_soundManager:playSample(spec.samples.activate)

                    if spec.toggleSpeedSyncEventId ~= nil then
                        g_inputBinding:setActionEventText(spec.toggleSpeedSyncEventId, g_i18n:getText("action_deactivateSpeedSync"), VehicleSpeedSync.MOD_NAME)
                    end
                else
                    if connectionLost == true then
                        g_soundManager:playSample(spec.samples.connectionLost)
                    else
                        g_soundManager:playSample(spec.samples.deactivate)
                    end

                    if spec.toggleSpeedSyncEventId ~= nil then
                        g_inputBinding:setActionEventText(spec.toggleSpeedSyncEventId, g_i18n:getText("action_activateSpeedSync"), VehicleSpeedSync.MOD_NAME)
                    end
                end
            end
        end
    end
end

function VehicleSpeedSync:getSyncedSpeedActive()
    local spec = self.spec_vehicleSpeedSync
    return spec ~= nil and spec.speedSyncActive and spec.vehicleInRange ~= nil
end

function VehicleSpeedSync:getSyncedSpeedWaiting()
    return self.spec_vehicleSpeedSync.isWaiting
end

function VehicleSpeedSync:getSyncedVehicleSpeed()
    if self:getSyncedSpeedActive() then
        local vehicleInRange = self.spec_vehicleSpeedSync.vehicleInRange
        local speed = vehicleInRange:getLastSpeed() * vehicleInRange.spec_motorized.speedDisplayScale

        if speed >= 0.5 then
            return math.floor(speed + 0.5)
        end
    end

    return 0
end

function VehicleSpeedSync:actionEventSpeedSyncState(actionName, inputValue, callbackState, isAnalog)
    local spec = self.spec_vehicleSpeedSync

    if spec ~= nil then
        if not spec.speedSyncActive then
            local vehicleInRange = self:findVehicleInRange(spec.setRange)

            if vehicleInRange ~= nil then
                self:setSyncedSpeedState(true, vehicleInRange)
            else
                g_currentMission:showBlinkingWarning(g_i18n:getText("warning_actionNotAllowedNow"))
            end
        else
            self:setSyncedSpeedState(false)
        end
    end
end

function VehicleSpeedSync:actionEventSyncCruiseControl(actionName, inputValue, callbackState, isAnalog)
    local vehicleInRange

    if self:getSyncedSpeedActive() then
        vehicleInRange = self.spec_vehicleSpeedSync.vehicleInRange
    else
        vehicleInRange = self:findVehicleInRange(self.spec_vehicleSpeedSync.setRange)
    end

    if vehicleInRange ~= nil then
        local speed = math.floor(vehicleInRange:getLastSpeed() + 0.5)

        if speed > 0 then
            local drivableSpec = self.spec_drivable

            self:setCruiseControlMaxSpeed(speed)

            if speed ~= drivableSpec.cruiseControl.speedSent then
                if g_server ~= nil then
                    g_server:broadcastEvent(SetCruiseControlSpeedEvent:new(self, speed), nil, nil, self)
                else
                    g_client:getServerConnection():sendEvent(SetCruiseControlSpeedEvent:new(self, speed))
                end

                drivableSpec.cruiseControl.speedSent = speed
            end

            self:setCruiseControlState(Drivable.CRUISECONTROL_STATE_ACTIVE)
        end
    else
        g_currentMission:showBlinkingWarning(g_i18n:getText("warning_actionNotAllowedNow"))
    end
end

function VehicleSpeedSync:updateVehiclePhysics(superFunc, axisForward, axisSide, doHandbrake, dt)
    local spec = self.spec_vehicleSpeedSync

    if spec ~= nil and spec.speedSyncActive then
        if axisForward == 0 then
            if self:getVehicleIsInRange(spec.vehicleInRange, spec.setRange) then
                local motor = self:getMotor()

                local speedLimit = motor:getSpeedLimit(true)
                local lastSpeed = spec.vehicleInRange:getLastSpeed()

                if lastSpeed > 0.5 and lastSpeed < speedLimit + 5 then
                    axisForward = MathUtil.sign(spec.vehicleInRange.movingDirection or 1)
                    motor:setSpeedLimit(math.min(lastSpeed, speedLimit))
                    spec.isWaiting = false
                elseif not spec.isWaiting then
                    self:setSyncedSpeedState(false, nil, true)
                end
            else
                self:setSyncedSpeedState(false, nil, true)
            end
        else
            self:setSyncedSpeedState(false)
        end
    end

    return superFunc(self, axisForward, axisSide, doHandbrake, dt)
end

function VehicleSpeedSync:getCanStartAIVehicle(superFunc)
    return superFunc(self) and not self:getSyncedSpeedActive()
end

function VehicleSpeedSync:setCruiseControlState(superFunc, state, noEventSend)
    if state ~= Drivable.CRUISECONTROL_STATE_OFF then
        self:setSyncedSpeedState(false, nil, false, true)
    end

    superFunc(self, state, noEventSend)
end

function VehicleSpeedSync:setCruiseControlMaxSpeed(superFunc, speed)
    if speed ~= nil then
        self:setSyncedSpeedState(false, nil, false, true)
    end

    superFunc(self, speed)
end
